"""
Query PyPI from the command line

``qypi`` is a command-line client for querying & searching `PyPI
<https://pypi.org>`_ for Python package information and outputting JSON (with
some minor opinionated changes to the output data structures).

Run ``qypi --help`` or visit <https://github.com/jwodder/qypi> for more
information.
"""

__version__      = '0.5.0'
__author__       = 'John Thorvald Wodder II'
__author_email__ = 'qypi@varonathe.org'
__license__      = 'MIT'
__url__          = 'https://github.com/jwodder/qypi'
